from enum import Enum, auto


class Sentinel(Enum):
    """
    Define a Sentinel type.

    See this for an explanation of the use-case for sentinels: https://peps.python.org/pep-0661/
    """

    MISSING = auto()
    NOT_GIVEN = auto()



from collections.abc import Callable
from functools import wraps
from typing import Any

import typer

from typer_repyt import build_command, DecDef


def complex_decorator(a: str, b: int) -> Callable[..., Any]:
    def _decorate(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            print(f"Complex decorator args: {a=}, {b=}")
            print("Complex decorator before function call")
            result = func(*args, **kwargs)
            print("Complex decorator after function call")
            return result
        return wrapper
    return _decorate


cli = typer.Typer()


@cli.command()
@complex_decorator("jawa", 13)
def static():
    print("In command")


def dynamic():
    print("In command")

build_command(
    cli,
    dynamic,
    decorators=[DecDef(dec_func=complex_decorator, dec_args=["jawa", 13], is_simple=False)],
)


if __name__ == "__main__":
    cli()



from collections.abc import Callable
from functools import wraps
from typing import Any, Annotated

import pytest
import typer

from typer_repyt.build_command import build_command, OptDef, ArgDef, ParamDef, DecDef
from typer_repyt.exceptions import BuildCommandError, RepytError

from tests.helpers import check_output, check_help, match_output, match_help


def simple_decorator(func: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        print("Start simple decorator")
        result = func(*args, **kwargs)
        print("End simple decorator")
        return result

    return wrapper


def complex_decorator(a: str, k: str = "hutt") -> Callable[..., Any]:
    def _decorate(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            print(f"Complex decorator args: {a=}, {k=}")
            print(f"Complex decorator before function call: {args=}, {kwargs=}")
            result = func(*args, **kwargs)
            print(f"Complex decorator after function call: {result=}")
            return result

        return wrapper

    return _decorate


def test_reference_static_implementation():
    cli = typer.Typer()

    @cli.command()
    @simple_decorator
    @complex_decorator("jawa", k="ewok")
    def static(  # pyright: ignore[reportUnusedFunction]
        ctx: typer.Context,  # pyright: ignore[reportUnusedParameter]
        mite1: Annotated[str, typer.Argument(help="This is mighty argument 1")],
        dyna2: Annotated[int, typer.Option(help="This is dynamic option 2")],
        dyna1: Annotated[str, typer.Option(help="This is dynamic option 1")] = "default1",
        mite2: Annotated[int | None, typer.Argument(help="This is mighty argument 2")] = None,
    ):
        """
        Just prints values of passed params
        """
        print(f"{dyna1=}, {dyna2=}, {mite1=}, {mite2=}")

    check_output(cli, "--dyna2=13", "17", expected_substring="dyna1='default1', dyna2=13")
    check_output(
        cli, "--dyna1=anyd", "--dyna2=13", "etim", expected_substring="dyna1='anyd', dyna2=13, mite1='etim', mite2=None"
    )
    match_output(cli, "etim", expected_pattern="Error.*Missing option '--dyna2'", exit_code=2)
    match_output(cli, "--dyna2=13", expected_pattern="Error.*Missing argument 'MITE1'", exit_code=2)

    match_output(
        cli,
        "--dyna2=13",
        "17",
        expected_pattern=[
            "Start simple decorator",
            "Complex decorator args: a='jawa', k='ewok'",
            r"Complex decorator before function call: args=\(\), kwargs={'ctx': .*Context.*, 'mite1': '17', 'dyna2': 13, 'dyna1': 'default1', 'mite2': None}",
            "dyna1='default1', dyna2=13",
            "Complex decorator after function call: result=None",
            "End simple decorator",
        ],
    )

    match_help(
        cli,
        expected_pattern=[
            "Just prints values of passed params",
            r"mite1 TEXT This is mighty argument 1 \[default:None\] \[required\]",
            r"mite2 \[MITE2\] This is mighty argument 2 \[default: None\]",
            r"--dyna2 INTEGER This is dynamic option2 \[default: None\] \[required\]",
            r"--dyna1 TEXT This is dynamic option 1 \[default: default1\]",
        ],
    )


def test_equivalent_dynamic_implementation():
    cli = typer.Typer()

    def dynamic(dyna1: str, dyna2: int, mite1: str, mite2: int | None):
        """
        Just prints values of passed params
        """
        print(f"{dyna1=}, {dyna2=}, {mite1=}, {mite2=}")

    build_command(
        cli,
        dynamic,
        OptDef(name="dyna1", param_type=str, help="This is dynamic option 1", default="default1"),
        OptDef(name="dyna2", param_type=int, help="This is dynamic option 2"),
        ArgDef(name="mite1", param_type=str, help="This is mighty argument 1"),
        ArgDef(name="mite2", param_type=int | None, help="This is mighty argument 2", default=None),
        decorators=[
            DecDef(simple_decorator),
            DecDef(complex_decorator, dec_args=["jawa"], dec_kwargs=dict(k="ewok"), is_simple=False),
        ],
        include_context=True,
    )

    check_output(cli, "--dyna2=13", "17", expected_substring="dyna1='default1', dyna2=13")
    check_output(
        cli, "--dyna1=anyd", "--dyna2=13", "etim", expected_substring="dyna1='anyd', dyna2=13, mite1='etim', mite2=None"
    )
    match_output(cli, "etim", expected_pattern="Error.*Missing option '--dyna2'", exit_code=2)
    match_output(cli, "--dyna2=13", expected_pattern="Error.*Missing argument 'MITE1'", exit_code=2)

    match_output(
        cli,
        "--dyna2=13",
        "17",
        expected_pattern=[
            "Start simple decorator",
            "Complex decorator args: a='jawa', k='ewok'",
            r"Complex decorator before function call: args=\(\), kwargs={'ctx': .*Context.*, 'dyna2': 13, 'mite1': '17', 'dyna1': 'default1', 'mite2': None}",
            "dyna1='default1', dyna2=13",
            "Complex decorator after function call: result=None",
            "End simple decorator",
        ],
    )

    match_help(
        cli,
        expected_pattern=[
            "Just prints values of passed params",
            r"mite1 TEXT This is mighty argument 1 \[default:None\] \[required\]",
            r"mite2 \[MITE2\] This is mighty argument 2 \[default: None\]",
            r"--dyna2 INTEGER This is dynamic option2 \[default: None\] \[required\]",
            r"--dyna1 TEXT This is dynamic option 1 \[default: default1\]",
        ],
    )


def test_build_command__option__with_no_help():
    cli = typer.Typer()

    def dynamic(dyna: str):
        print(f"{dyna=}")

    build_command(
        cli,
        dynamic,
        OptDef(name="dyna", param_type=str),
    )

    check_output(cli, "--dyna=ZOOM", expected_substring="dyna='ZOOM'")
    match_help(cli, expected_pattern=r"--dyna TEXT \[default: None\] \[required\]")


def test_build_command__option__rich_help_panel():
    cli = typer.Typer()

    def dynamic(dyna: str):
        print(f"{dyna=}")

    build_command(
        cli,
        dynamic,
        OptDef(name="dyna", param_type=str, rich_help_panel="This is a dynamic option"),
    )

    check_output(cli, "--dyna=ZOOM", expected_substring="dyna='ZOOM'")
    match_help(cli, expected_pattern=r"This is a dynamic option.*--dyna TEXT \[default: None\] \[required\]")


def test_build_command__option__no_show_default():
    cli = typer.Typer()

    def dynamic(dyna: str):
        print(f"{dyna=}")

    build_command(
        cli,
        dynamic,
        OptDef(name="dyna", param_type=str, default="ZOOM", show_default=False),
    )

    check_output(cli, expected_substring="dyna='ZOOM'")
    match_help(cli, expected_pattern=r"--dyna TEXT(?! \[default)")


def test_build_command__option__prompt__default():
    cli = typer.Typer()

    def dynamic(dyna: str):
        print(f"{dyna=}")

    build_command(
        cli,
        dynamic,
        OptDef(name="dyna", param_type=str, prompt=True),
    )

    check_output(cli, expected_substring="dyna='ZOOM'", input="ZOOM")
    match_help(cli, expected_pattern=r"--dyna TEXT \[default: None\] \[required\]")


def test_build_command__option__prompt__custom():
    cli = typer.Typer()

    def dynamic(dyna: str):
        print(f"{dyna=}")

    build_command(
        cli,
        dynamic,
        OptDef(name="dyna", param_type=str, prompt="Gimme an answer"),
    )

    check_output(cli, expected_substring="dyna='ZOOM'", input="ZOOM\n")
    match_help(cli, expected_pattern=r"--dyna TEXT \[default: None\] \[required\]")


# TODO: Figure out how to fix this test
@pytest.mark.skip("Passing input for confirmation prompt is not working")
def test_build_command__option__confirmation_prompt():
    cli = typer.Typer()

    def dynamic(dyna: str):
        print(f"{dyna=}")

    build_command(
        cli,
        dynamic,
        OptDef(name="dyna", param_type=str, prompt=True, confirmation_prompt=True),
    )

    check_output(cli, expected_substring=["Dyna: ZOOM", "dyna='ZOOM'"], input="ZOOM\nBOOM\n\n\n")
    check_help(cli, expected_pattern=r"--dyna TEXT \[default: None\] \[required\]")


def test_build_command__option__hide_input():
    cli = typer.Typer()

    def dynamic(dyna: str):
        print(f"{dyna=}")

    build_command(
        cli,
        dynamic,
        OptDef(name="dyna", param_type=str, prompt=True, hide_input=True),
    )

    check_output(cli, expected_substring="dyna='ZOOM'", input="ZOOM\n")
    match_help(cli, expected_pattern=r"--dyna TEXT \[default: None\] \[required\]")


def test_build_command__option__override_name():
    cli = typer.Typer()

    def dynamic(dyna: str):
        print(f"{dyna=}")

    build_command(
        cli,
        dynamic,
        OptDef(name="dyna", param_type=str, override_name="dyyyna"),
    )

    check_output(cli, "--dyyyna=ZOOM", expected_substring="dyna='ZOOM'")
    match_help(cli, expected_pattern=r"--dyyyna TEXT \[default: None\] \[required\]")


def test_build_command__option__short_name():
    cli = typer.Typer()

    def dynamic(dyna: str):
        print(f"{dyna=}")

    build_command(
        cli,
        dynamic,
        OptDef(name="dyna", param_type=str, short_name="d"),
    )

    check_output(cli, "-dZOOM", expected_substring="dyna='ZOOM'")

    # TODO: Maybe create an issue for this on typer github? Seems like the options should 1000% be stripped
    # check_output(cli, "-d    ZOOM", expected_substring="dyna='ZOOM'")

    # TODO: figure out what the star in the help output means
    # dyna_pattern = "\* -a TEXT [default: None] [required]"
    match_help(cli, expected_pattern=r"-d TEXT \[default: None\] \[required\]")


def test_build_command__option__callback():
    cli = typer.Typer()

    def back(val: Any):
        print(f"{val=}")
        return val

    def dynamic(dyna: str):
        print(f"{dyna=}")

    build_command(
        cli,
        dynamic,
        OptDef(name="dyna", param_type=str, callback=back),
    )

    check_output(cli, "--dyna=ZOOM", expected_substring=["dyna='ZOOM'", "val='ZOOM'"])


def test_build_command__option__is_eager():
    cli = typer.Typer()

    def back1(val: Any):
        print(f"back1: {val=}")
        raise typer.Exit()

    def back2(val: Any):
        print(f"back2: {val=}")
        raise typer.Exit()

    def dynamic(dyna: str):
        print(f"{dyna=}")

    build_command(
        cli,
        dynamic,
        OptDef(name="dyna", param_type=str, callback=back1),
        OptDef(name="mite", param_type=str, callback=back2, is_eager=True),
    )

    check_output(cli, "--dyna=ZOOM", "--mite=BOOM", expected_substring="back2: val='BOOM'")


def test_build_command__argument__with_no_help():
    cli = typer.Typer()

    def dynamic(mite: str):
        print(f"{mite=}")

    build_command(
        cli,
        dynamic,
        ArgDef(name="mite", param_type=str),
    )

    check_output(cli, "BOOM", expected_substring="mite='BOOM'")
    match_help(cli, expected_pattern=r"mite TEXT \[default: None\] \[required\]")


def test_build_command__argument__rich_help_panel():
    cli = typer.Typer()

    def dynamic(mite: str):
        print(f"{mite=}")

    build_command(
        cli,
        dynamic,
        ArgDef(name="mite", param_type=str, rich_help_panel="This is a mighty argument"),
    )

    check_output(cli, "BOOM", expected_substring="mite='BOOM'")
    match_help(cli, expected_pattern=r"This is a mighty argument.*mite TEXT \[default: None\] \[required\]")


def test_build_command__argument__no_show_default():
    cli = typer.Typer()

    def dynamic(mite: str):
        print(f"{mite=}")

    build_command(
        cli,
        dynamic,
        ArgDef(name="mite", param_type=str, default="BOOM", show_default=False),
    )

    check_output(cli, expected_substring="mite='BOOM'")
    match_help(cli, expected_pattern=r"mite \[MITE\](?! \[default)")


def test_build_command__argument__metavar():
    cli = typer.Typer()

    def dynamic(mite: str):
        print(f"{mite=}")

    build_command(
        cli,
        dynamic,
        ArgDef(name="mite", param_type=str, metavar="NITRO"),
    )

    check_output(cli, "BOOM", expected_substring="mite='BOOM'")
    match_help(cli, expected_pattern=r"mite NITRO \[default: None\] \[required\]")


def test_build_command__argument__hidden():
    cli = typer.Typer()

    def dynamic(mite: str):
        print(f"{mite=}")

    build_command(
        cli,
        dynamic,
        ArgDef(name="mite", param_type=str, hidden=True),
    )

    match_output(cli, expected_pattern="Error.*Missing argument 'MITE'", exit_code=2)
    check_output(cli, "BOOM", expected_substring="mite='BOOM'")
    check_output(cli, "--help", expected_substring=None)


def test_build_command__argument__envvar__single():
    cli = typer.Typer()

    def dynamic(mite: str):
        print(f"{mite=}")

    build_command(
        cli,
        dynamic,
        ArgDef(name="mite", param_type=str, envvar="MITE"),
    )

    match_output(cli, expected_pattern="Error.*Missing argument 'MITE'", exit_code=2)
    check_output(cli, expected_substring="mite='BOOM'", env_vars=dict(MITE="BOOM"))
    match_help(cli, expected_pattern=r"mite TEXT \[env var: MITE\] \[default: None\] \[required\]")


def test_build_command__argument__envvar__multiple():
    cli = typer.Typer()

    def dynamic(mite: str):
        print(f"{mite=}")

    build_command(
        cli,
        dynamic,
        ArgDef(name="mite", param_type=str, envvar=["MITE", "NITRO"]),
    )

    match_output(cli, expected_pattern="Error.*Missing argument 'MITE'", exit_code=2)
    check_output(cli, expected_substring="mite='BOOM'", env_vars=dict(MITE="BOOM"))
    check_output(cli, expected_substring="mite='BOOM'", env_vars=dict(NITRO="BOOM"))
    match_help(cli, expected_pattern=r"mite TEXT \[env var: MITE, NITRO\] \[default: None\] \[required\]")


def test_build_command__argument__showenvvar():
    cli = typer.Typer()

    def dynamic(mite: str):
        print(f"{mite=}")

    build_command(
        cli,
        dynamic,
        ArgDef(name="mite", param_type=str, envvar=["MITE"], show_envvar=False),
    )

    check_output(cli, expected_substring="mite='BOOM'", env_vars=dict(MITE="BOOM"))
    match_help(cli, expected_pattern=r"mite TEXT(?! \[env var: MITE\]) \[default: None\] \[required\]")


def test_build_command__decorator__simple():
    cli = typer.Typer()

    def dynamic(dyna: str):
        print(f"{dyna=}")

    build_command(
        cli,
        dynamic,
        OptDef(name="dyna", param_type=str),
        decorators=[DecDef(simple_decorator)],
    )

    match_output(
        cli,
        "--dyna=ZOOM",
        expected_pattern=[
            "Start simple decorator",
            "dyna='ZOOM'",
            "End simple decorator",
        ],
    )


def test_build_command__decorator__complex():
    cli = typer.Typer()

    def dynamic(dyna: str):
        print(f"{dyna=}")

    build_command(
        cli,
        dynamic,
        OptDef(name="dyna", param_type=str),
        decorators=[DecDef(complex_decorator, dec_args=["jawa"], dec_kwargs=dict(k="ewok"), is_simple=False)],
    )

    match_output(
        cli,
        "--dyna=ZOOM",
        expected_pattern=[
            "Complex decorator args: a='jawa', k='ewok'",
            r"Complex decorator before function call: args=\(\), kwargs={'dyna': 'ZOOM'}",
            "dyna='ZOOM'",
            "Complex decorator after function call: result=None",
        ],
    )


def test_build_command__decorator__multiple():
    cli = typer.Typer()

    def dynamic(dyna: str):
        print(f"{dyna=}")

    build_command(
        cli,
        dynamic,
        OptDef(name="dyna", param_type=str),
        decorators=[
            DecDef(simple_decorator),
            DecDef(complex_decorator, dec_args=["jawa"], dec_kwargs=dict(k="ewok"), is_simple=False),
        ],
    )

    match_output(
        cli,
        "--dyna=ZOOM",
        expected_pattern=[
            "Start simple decorator",
            "Complex decorator args: a='jawa', k='ewok'",
            r"Complex decorator before function call: args=\(\), kwargs={'dyna': 'ZOOM'}",
            "dyna='ZOOM'",
            "Complex decorator after function call: result=None",
            "End simple decorator",
        ],
    )


def test_build_command__decorator__raises_exception_if_args_set_for_simple_decorator():
    cli = typer.Typer()

    def dynamic(dyna: str):
        print(f"{dyna=}")

    with pytest.raises(BuildCommandError, match="Decorator arguments are not allowed for simple decorators"):
        build_command(
            cli,
            dynamic,
            decorators=[
                DecDef(simple_decorator, dec_args=["jawa"], dec_kwargs=dict(k="ewok")),
            ],
        )


def test_build_command__raises_error_on_unsupported_param_def():
    cli = typer.Typer()

    def dynamic(dyna: str):
        print(f"{dyna=}")

    with pytest.raises(RepytError, match="Unsupported parameter definition type"):
        build_command(cli, dynamic, ParamDef(name="mite", param_type=str))



from typing import Annotated

from typer import Typer, Option
from typer_repyt import build_command, OptDef

cli = Typer()


def back1(val: str):
    print(f"Callback 1 operating on {val=}")
    return f"one: {val}"

def back2(val: str):
    print(f"Callback 2 operating on {val=}")
    return f"two: {val}"


@cli.command()
def static(
    dyna1: Annotated[str, Option(callback=back1)],
    dyna2: Annotated[str, Option(callback=back2, is_eager=True)],
):
    print(f"{dyna1=}, {dyna2=}")


def dynamic(dyna1: str, dyna2: str):
    print(f"{dyna1=}, {dyna2=}")


build_command(
    cli,
    dynamic,
    OptDef(name="dyna1", param_type=str, callback=back1),
    OptDef(name="dyna2", param_type=str, callback=back2, is_eager=True),
)


if __name__ == "__main__":
    cli()



import ast
from functools import reduce, update_wrapper
import inspect
import textwrap
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import Any, cast, Literal
from types import UnionType

from inflection import dasherize
import typer

from typer_repyt.constants import Sentinel
from typer_repyt.exceptions import BuildCommandError


@dataclass
class ParamDef:
    """
    Define the necessary components to build a Typer `Option` or `Argument`.

    These elements are used by both `OptDef` and `ArgDef`.
    """

    name: str
    param_type: UnionType | type[Any]
    default: Any | None | Literal[Sentinel.NOT_GIVEN] = Sentinel.NOT_GIVEN
    help: str | None = None
    rich_help_panel: str | None = None
    show_default: bool | str = True


@dataclass
class OptDef(ParamDef):
    """
    Define the additional components to build a Typer `Option`.
    """

    prompt: bool | str = False
    confirmation_prompt: bool = False
    hide_input: bool = False
    override_name: str | None = None
    short_name: str | None = None
    callback: Callable[..., Any] | None = None
    is_eager: bool = False


@dataclass
class ArgDef(ParamDef):
    """
    Define the additional components to build a Typer `Argument`.
    """

    metavar: str | None = None
    hidden: bool = False
    envvar: str | list[str] | None = None
    show_envvar: bool = True


@dataclass
class DecDef:
    """
    Define a decorator function and it parameters.
    """

    dec_func: Callable[..., Any]
    dec_args: list[Any] = field(default_factory=list)
    dec_kwargs: dict[str, Any] = field(default_factory=dict)
    is_simple: bool = True

    def decorate(self, f: Callable[..., Any]) -> Callable[..., Any]:
        BuildCommandError.require_condition(
            not self.is_simple or (self.dec_args == [] and self.dec_kwargs == {}),
            "Decorator arguments are not allowed for simple decorators",
        )

        def wrap(*args: Any, **kwargs: Any) -> Any:
            if self.is_simple:
                return self.dec_func(f)(*args, **kwargs)
            else:
                return self.dec_func(*self.dec_args, **self.dec_kwargs)(f)(*args, **kwargs)

        update_wrapper(wrap, f)
        return wrap


def build_command(
    cli: typer.Typer,
    func: Callable[..., None],
    /,
    *param_defs: ParamDef,
    decorators: list[DecDef] | None = None,
    include_context: bool | str = False,  # TODO: Think about setting this automatically if func's first arg is ctx
):
    """
    Build a Typer command dynamically based on a function template and list of argument definitions.

    Args:
        cli:             The Typer app that the command should be added to
        func:            A "template" function that will be used to build out the final function.
                         The name of the function will be preserved as will its docstring.
                         Though it is useful to define arguments for the function that match the opt_defs, it is not necessary.
                         It will, however, help with static type checking.
        param_defs:      Argument definitions that will be used to dynamically build the Typer command.
        decorators:      An optional list of decorators to apply to the command function. Like regular decorators, they are
                         applied in reverse order with those nearest to the function definition being called first.
        include_context: If set, include the `typer.Context` as the first argument to the function. If the value is not
                         boolean, use the passed string as the name of the context arg.

    The following two command definitions are equivalent:

    ```python
    cli = typer.Typer()

    @cli.command()
    @simple_decorator
    @complex_decorator("nitro", glycerine=True)
    def static(
        t_ctx: Typer.Context,
        mite1: Annotated[str, typer.Argument(help="This is mighty argument 1")],
        dyna2: Annotated[int, typer.Option(help="This is dynamic option 2")],
        dyna1: Annotated[str, typer.Option(help="This is dynamic option 1")] = "default1",
        mite2: Annotated[int | None, typer.Argument(help="This is mighty argument 2")] = None,
    ):
        '''
        Just prints values of passed params
        '''
        print(f"{dyna1=}, {dyna2=}, {mite1=}, {mite2=}")


    def dynamic(dyna1: str, dyna2: int, mite1: str, mite2: int | None):
        '''
        Just prints values of passed params
        '''
        print(f"{dyna1=}, {dyna2=}, {mite1=}, {mite2=}")


    build_command(
        cli,
        dynamic,
        OptDef(name="dyna1", param_type=str, help="This is dynamic option 1", default="default1"),
        OptDef(name="dyna2", param_type=int, help="This is dynamic option 2"),
        ArgDef(name="mite1", param_type=str, help="This is mighty argument 1"),
        ArgDef(name="mite2", param_type=int | None, help="This is mighty argument 2", default=None),
        decorators=[
            DecDef(simple_decorator),
            DecDef(complex_decorator, dec_args=["nitro"], dec_kwargs=dict(glycerine=True)),
        ],
        include_context="t_ctx",
    )

    ```
    """
    # TODO: Add some examples to the docstring

    # These will hold the args and kwargs (with defaults) for the final constructed function
    args: list[ast.arg] = []
    kwonlyargs: list[ast.arg] = []
    kw_defaults: list[Any] = []

    # Create a local namespace for the compiled function
    namespace: dict[str, Any] = {}

    # Convert each ParamDef into an Annotated keyword argument for the constructed function
    for param_def in param_defs:
        param_args: list[Any] = []

        keywords: list[ast.keyword] = [
            ast.keyword(arg="help", value=ast.Constant(value=param_def.help)),
            ast.keyword(arg="rich_help_panel", value=ast.Constant(value=param_def.rich_help_panel)),
            ast.keyword(arg="show_default", value=ast.Constant(value=param_def.show_default)),
        ]

        # If the ParamDef is an OptDef, assemble args and keywords for it
        if isinstance(param_def, OptDef):
            if param_def.override_name:
                param_args.append(ast.Constant(value=f"--{dasherize(param_def.override_name)}"))
            if param_def.short_name:
                param_args.append(ast.Constant(value=f"-{param_def.short_name}"))
            if param_def.callback:
                keywords.append(ast.keyword(arg="callback", value=ast.Name(param_def.callback.__name__)))
                namespace[param_def.callback.__name__] = param_def.callback
            keywords.extend(
                [
                    ast.keyword(arg="prompt", value=ast.Constant(value=param_def.prompt)),
                    ast.keyword(arg="confirmation_prompt", value=ast.Constant(value=param_def.confirmation_prompt)),
                    ast.keyword(arg="hide_input", value=ast.Constant(value=param_def.hide_input)),
                    ast.keyword(arg="is_eager", value=ast.Constant(value=param_def.is_eager)),
                ]
            )

        # If the ParamDef is an ArgDef, assemble args and keywords for it
        elif isinstance(param_def, ArgDef):
            keywords.extend(
                [
                    ast.keyword(arg="metavar", value=ast.Constant(value=param_def.metavar)),
                    ast.keyword(arg="hidden", value=ast.Constant(value=param_def.hidden)),
                    ast.keyword(arg="show_envvar", value=ast.Constant(value=param_def.show_envvar)),
                ]
            )

            if isinstance(param_def.envvar, str):
                keywords.append(ast.keyword(arg="envvar", value=ast.Constant(value=param_def.envvar)))
            elif isinstance(param_def.envvar, list):
                keywords.append(
                    ast.keyword(
                        arg="envvar",
                        value=ast.List(elts=[ast.Constant(value=env) for env in param_def.envvar]),
                    )
                )

        # If the ParamDef is not an OptDef or ArgDef, raise an error indicating that it's unsupported
        else:
            raise BuildCommandError(f"Unsupported parameter definition type: {type(param_def)}")

        # Get the type annotation for the opt/arg
        # Just found out that I didn't need to do this (yet) because Typer does not support Union types
        # I will keep it here in case it does soon: https://github.com/fastapi/typer/pull/1148
        type_expr: ast.Name | ast.BinOp
        if isinstance(param_def.param_type, UnionType):

            def bitor(left: ast.expr | type[Any], right: type[Any]) -> ast.BinOp:
                if not isinstance(left, ast.expr):
                    left = ast.Name(id=left.__name__)
                return ast.BinOp(left=left, op=ast.BitOr(), right=ast.Name(id=right.__name__))

            type_expr = reduce(bitor, param_def.param_type.__args__)
        else:
            type_expr = ast.Name(id=param_def.param_type.__name__)

        # Create the actual Annotated element using func, args, and keywords assembled above
        param_attr = "Option" if isinstance(param_def, OptDef) else "Argument"
        annotation = ast.Subscript(
            value=ast.Name(id="Annotated"),
            slice=ast.Tuple(
                elts=[
                    type_expr,
                    ast.Call(
                        func=ast.Attribute(value=ast.Name(id="typer"), attr=param_attr),
                        args=param_args,
                        keywords=keywords,
                    ),
                ],
            ),
        )

        # Add the constructed CLI option/argument.
        # If a default is not provided, it is added as an argument. Otherwise, it is added as a kwonlyarg
        arg = ast.arg(arg=param_def.name, annotation=annotation)
        if param_def.default is Sentinel.NOT_GIVEN:
            args.append(arg)
        else:
            kwonlyargs.append(arg)
            kw_defaults.append(ast.Constant(value=param_def.default))

    # Extract the function body from the template function
    source = textwrap.dedent(inspect.getsource(func)).strip()

    parsed_ast: ast.Module = ast.parse(source)
    BuildCommandError.require_condition(
        len(parsed_ast.body) == 1, "Function template must contain exactly one function definition"
    )
    fdef: ast.FunctionDef = cast(ast.FunctionDef, parsed_ast.body[0])
    body = fdef.body

    # Add imports for Annotated, typer.Option, and NoneType
    imports: list[ast.Import | ast.ImportFrom] = [
        ast.ImportFrom(module="typing", names=[ast.alias(name="Annotated")], level=0),
        ast.Import(names=[ast.alias(name="typer")]),
        ast.ImportFrom(module="types", names=[ast.alias(name="NoneType")], level=0),
        ast.ImportFrom(module="typer", names=[ast.alias(name="Context")], level=0),
    ]

    # Add the typer context if requested
    # TODO: Test this
    if include_context:
        ctx_name = "ctx" if isinstance(include_context, bool) else include_context
        ctx_arg = ast.arg(arg=ctx_name, annotation=ast.Name(id="Context"))
        args.insert(0, ctx_arg)

    # Create the function definition
    func_def = ast.FunctionDef(
        name=func.__name__,
        args=ast.arguments(
            args=args,
            kwonlyargs=kwonlyargs,
            kw_defaults=kw_defaults,
        ),
        body=body,
    )

    # Compile a module containing the imports and function definition
    module = ast.Module(body=[*imports, func_def])
    _set_ast_location(module)
    code = compile(module, filename="<ast>", mode="exec")

    # execute the compiled module
    exec(code, func.__globals__, namespace)

    command_func = namespace[func.__name__]

    # Add decorators to the command function
    if decorators is not None:
        with BuildCommandError.handle_errors("There was a problem adding decorators"):
            for dec_def in reversed(decorators):
                command_func = dec_def.decorate(command_func)

    # Add the command function to the Typer app
    cli.command()(command_func)


def _set_ast_location(node: ast.AST, lineno: int = 0, col_offset: int = 0):
    """
    Set the `lineno` and `col_offset` attributes for all nodes (that need them).

    This is necessary, because the constructed AST from `build_command` requires many nodes to have the `lineno` and
    `col_offset` values set. Even though these are not used for anything, it's important to set these values to produce
    a valid AST.
    """
    for child in ast.iter_child_nodes(node):
        _set_ast_location(child, lineno, col_offset)
    if "lineno" in node._attributes:
        setattr(node, "lineno", lineno)
    if "col_offset" in node._attributes:
        setattr(node, "col_offset", col_offset)



from typing import Annotated

from typer import Typer, Option
from typer_repyt import build_command, OptDef

cli = Typer()


def back(dyna: str):
    print(f"Callback operating on {dyna=}")
    return dyna * 3


@cli.command()
def static(dyna: Annotated[str, Option(callback=back)]):
    print(f"{dyna=}")


def dynamic(dyna: str):
    print(f"{dyna=}")


build_command(
    cli,
    dynamic,
    OptDef(name="dyna", param_type=str, callback=back),
)


if __name__ == "__main__":
    cli()



"""
This set of demos shows the use of the `attach_settings` decorator.
"""

from collections.abc import Callable
from functools import wraps
from typing import Any
import typer
from typer_repyt import build_command, OptDef, ArgDef, DecDef


def demo_1__build_command__basic():
    """
    This function demonstrates the use of the `build_command()` function.
    The `build_command()` function allows you to dynamically build a Typer
    command using an existing function and some parameter definitions.
    """

    cli = typer.Typer()

    def dynamic(dyna1: str, dyna2: int, mite1: str, mite2: int | None):
        """
        Just prints values of passed params
        """
        print(f"{dyna1=}, {dyna2=}, {mite1=}, {mite2=}")

    build_command(
        cli,
        dynamic,
        OptDef(name="dyna1", param_type=str, help="This is dynamic option 1", default="default1"),
        OptDef(name="dyna2", param_type=int, help="This is dynamic option 2"),
        ArgDef(name="mite1", param_type=str, help="This is mighty argument 1"),
        ArgDef(name="mite2", param_type=int | None, help="This is mighty argument 2", default=None),
    )

    cli(["BOOM", "--dyna2=13"])


def demo_2__build_command__with_simple_decorator():
    """
    This function demonstrates how to use the `build_command()` function
    with a simple decorator. The decorator used in this demo is a basic
    decorator that would be used like this in non-dynamic context:

    ```python
    @cli.command()
    @simple_decorator
    def static(dyna: str):
        print(f"{dyna=}")
    ```
    """

    cli = typer.Typer()

    def simple_decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            print(f"Decorator before function call: {args=}, {kwargs=}")
            result = func(*args, **kwargs)
            print(f"Decorator after function call: {result=}")
            return result

        return wrapper

    def dynamic(dyna: str):
        """
        Just prints values of the passed param
        """
        print(f"{dyna=}")

    build_command(
        cli,
        dynamic,
        OptDef(name="dyna", param_type=str),
        decorators=[DecDef(dec_func=simple_decorator)],
    )

    cli(["--dyna=ZOOM"])


def demo_3__build_command__with_chained_decorators():
    """
    This function demonstrates how to use the `build_command()` function
    with multiple decorators. The decorators used in this demo are both
    basic. They are called in _reverse_ order of their definition with
    the decorator declared closer to the function being called first.
    The equivalent static definition would look like:

    ```python
    @cli.command()
    @first_decorator
    @second_decorator
    def static(dyna: str):
        print(f"{dyna=}")
    ```
    """

    cli = typer.Typer()

    def first_decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            print("Start first decorator")
            result = func(*args, **kwargs)
            print("End first decorator")
            return result

        return wrapper

    def second_decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            print("Start second decorator")
            result = func(*args, **kwargs)
            print("End second decorator")
            return result

        return wrapper

    def dynamic(dyna: str):
        """
        Just prints values of the passed param
        """
        print(f"{dyna=}")

    build_command(
        cli,
        dynamic,
        OptDef(name="dyna", param_type=str, help="This is a dynamic option", default="default1"),
        decorators=[DecDef(dec_func=first_decorator), DecDef(dec_func=second_decorator)],
    )

    cli(["--dyna=ZOOM"])


def demo_4__build_command__with_complex_decorator():
    """
    This function demonstrates how to use the `build_command()` function
    with a more complex decorator. The decorator used in this demo is takes
    its own `args` and `kwargs`. Here is the static equivalent:

    ```python
    @cli.command()
    @complex_decorator("jawa", k="ewok")
    def static(dyna: str):
        print(f"{dyna=}")
    ```
    """

    cli = typer.Typer()

    def complex_decorator(a: str, k: str = "hutt") -> Callable[..., Any]:
        def _decorate(func: Callable[..., Any]) -> Callable[..., Any]:
            @wraps(func)
            def wrapper(*args: Any, **kwargs: Any) -> Any:
                print(f"Decorator args: {a=}, {k=}")
                print(f"Decorator before function call: {args=}, {kwargs=}")
                result = func(*args, **kwargs)
                print(f"Decorator after function call: {result=}")
                return result

            return wrapper

        return _decorate

    def dynamic(dyna: str):
        """
        Just prints values of the passed param
        """
        print(f"{dyna=}")

    build_command(
        cli,
        dynamic,
        OptDef(name="dyna", param_type=str),
        decorators=[DecDef(dec_func=complex_decorator, dec_args=["jawa"], dec_kwargs={"k": "ewok"}, is_simple=False)],
    )

    cli(["--help"])



from typing import Annotated

from typer import Typer, Option
from typer_repyt import build_command, OptDef

cli = Typer()

@cli.command()
def static(class_name: Annotated[str, Option("--class")]):
    print(f"class={class_name}")


def dynamic(class_name: str):
    print(f"class={class_name}")


build_command(
    cli,
    dynamic,
    OptDef(name="class_name", param_type=str, override_name="class"),
)


if __name__ == "__main__":
    cli()



from typing import Annotated

from typer import Typer, Option
from typer_repyt import build_command, OptDef

cli = Typer()

@cli.command()
def static(
    dyna1: Annotated[str, Option(show_default=False)] = "BOOM",
    dyna2: Annotated[str, Option(show_default="-hidden-")] = "BOOM",
):
    print(f"{dyna1=}, {dyna2=}")


def dynamic(dyna1: str, dyna2: str):
    print(f"{dyna1=}, {dyna2=}")


build_command(
    cli,
    dynamic,
    OptDef(name="dyna1", param_type=str, default="BOOM", show_default=False),
    OptDef(name="dyna2", param_type=str, default="BOOM", show_default="-hidden-"),
)


if __name__ == "__main__":
    cli()



from collections.abc import Callable
from functools import wraps
from typing import Annotated, Any

import typer


def simple_decorator(func: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        print("Start simple decorator")
        result = func(*args, **kwargs)
        print("End simple decorator")
        return result
    return wrapper


def complex_decorator(a: str, k: str = "hutt") -> Callable[..., Any]:
    def _decorate(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            print(f"Complex decorator args: {a=}, {k=}")
            print(f"Complex decorator before function call: {args=}, {kwargs=}")
            result = func(*args, **kwargs)
            print(f"Complex decorator after function call: {result=}")
            return result
        return wrapper
    return _decorate


cli = typer.Typer()

@cli.command()
@simple_decorator
@complex_decorator("jawa", k="ewok")
def static(
    ctx: typer.Context,  # pyright: ignore[reportUnusedParameter]
    mite1: Annotated[str, typer.Argument(help="This is mighty argument 1")],
    dyna2: Annotated[int, typer.Option(help="This is dynamic option 2")],
    dyna1: Annotated[str, typer.Option(help="This is dynamic option 1")] = "default1",
    mite2: Annotated[int | None, typer.Argument(help="This is mighty argument 2")] = None,
):
    """
    Just prints values of passed params
    """
    print(f"{dyna1=}, {dyna2=}, {mite1=}, {mite2=}")


if __name__ == "__main__":
    cli()



from typing import Annotated

from typer import Typer, Option
from typer_repyt import build_command, OptDef

cli = Typer()

@cli.command()
def static(
    dyna1: Annotated[str, Option(prompt=True)],
    dyna2: Annotated[str, Option(prompt="Dyna2 goes")] = "POW",
):
    print(f"{dyna1=}, {dyna2=}")


def dynamic(dyna1: str, dyna2: str):
    print(f"{dyna1=}, {dyna2=}")


build_command(
    cli,
    dynamic,
    OptDef(name="dyna1", param_type=str, prompt=True),
    OptDef(name="dyna2", param_type=str, prompt="Dyna2 goes", default="POW"),
)


if __name__ == "__main__":
    cli()



from typing import Any
from collections.abc import Callable
from functools import wraps

from typer import Typer
from typer_repyt import build_command, DecDef

def simple_decorator(func: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        print("Start simple decorator")
        result = func(*args, **kwargs)
        print("End simple decorator")
        return result
    return wrapper

cli = Typer()

@cli.command()
@simple_decorator
def static():
    print("In command")


def dynamic():
    print("In command")

build_command(
    cli,
    dynamic,
    decorators=[DecDef(dec_func=simple_decorator)],
)


if __name__ == "__main__":
    cli()



from typing import Annotated

from typer import Typer, Argument
from typer_repyt import build_command, ArgDef

cli = Typer()


@cli.command()
def static(mite: Annotated[str, Argument(metavar="NITRO")]):
    print(f"{mite=}")


def dynamic(mite: str):
    print(f"{mite=}")


build_command(
    cli,
    dynamic,
    ArgDef(name="mite", param_type=str, metavar="NITRO"),
)


if __name__ == "__main__":
    cli()



from typing import Any
from collections.abc import Callable
from functools import wraps

import typer
from typer_repyt import build_command, OptDef, ArgDef, DecDef


def simple_decorator(func: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        print("Start simple decorator")
        result = func(*args, **kwargs)
        print("End simple decorator")
        return result
    return wrapper


def complex_decorator(a: str, k: str = "hutt") -> Callable[..., Any]:
    def _decorate(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            print(f"Complex decorator args: {a=}, {k=}")
            print(f"Complex decorator before function call: {args=}, {kwargs=}")
            result = func(*args, **kwargs)
            print(f"Complex decorator after function call: {result=}")
            return result
        return wrapper
    return _decorate


cli = typer.Typer()


def dynamic(ctx: typer.Context, dyna1: str, dyna2: int, mite1: str, mite2: int | None):  # pyright: ignore[reportUnusedParameter]
    """
    Just prints values of passed params
    """
    print(f"{dyna1=}, {dyna2=}, {mite1=}, {mite2=}")


build_command(
    cli,
    dynamic,
    OptDef(name="dyna1", param_type=str, help="This is dynamic option 1", default="default1"),
    OptDef(name="dyna2", param_type=int, help="This is dynamic option 2"),
    ArgDef(name="mite1", param_type=str, help="This is mighty argument 1"),
    ArgDef(name="mite2", param_type=int | None, help="This is mighty argument 2", default=None),
    decorators=[
        DecDef(simple_decorator),
        DecDef(complex_decorator, dec_args=["jawa"], dec_kwargs=dict(k="ewok"), is_simple=False),
    ],
    include_context=True,
)

if __name__ == "__main__":
    cli()



from typing import Annotated

from typer import Typer, Option
from typer_repyt import build_command, OptDef

cli = Typer()

@cli.command()
def static(dyna: Annotated[int | None, Option()]):
    print(f"{dyna=}")


def dynamic(dyna: int | None):
    print(f"{dyna=}")


build_command(cli, dynamic, OptDef(name="dyna", param_type=int | None))


if __name__ == "__main__":
    cli()



from typing import Annotated

from typer import Typer, Argument
from typer_repyt import build_command, ArgDef

cli = Typer()


@cli.command()
def static(
    mite1: Annotated[str, Argument(envvar="MITE")],
    mite2: Annotated[str, Argument(envvar=["NITRO", "DYNA", "MITE"])],
):
    print(f"{mite1=}, {mite2=}")


def dynamic(mite1: str, mite2: str):
    print(f"{mite1=}, {mite2=}")


build_command(
    cli,
    dynamic,
    ArgDef(name="mite1", param_type=str, envvar="MITE"),
    ArgDef(name="mite2", param_type=str, envvar=["NITRO", "DYNA", "MITE"]),
)


if __name__ == "__main__":
    cli()
